
// Minimal implementation of Gigabit Ethernet over Fiber 
// using the SerDes of ECP5UM-type FPGAs and the open-source yosys/nextpnr tool chain
//
// version 20220106, Pieter-Tjerk de Boer, pa3fwm@pa3fwm.nl
//
// Note that this is a minimal proof-of-concept implementation, although it does work,
// at least in my setup talking to an RTL8213-based media converter.
// Many aspects of the SerDes are still unclear to me, so I may well be doing things "wrongly".
// Many details of the full IEEE802 spec are not implemented, relying on the "other side" to do things right...
//
// Freely distributable under GPLv2


`include "crc32_d8.v"

module serdes_gbe (
       clk,    // input clock 125 MHz

       rxbyte,
       rxclk,
       rxdv,   // 1 if rxbyte is valid at the current rxclk posedge; when rxdv goes to 0, the frame has ended
       rxfin,  // 1 for 1 cycle at the end of a correctly-terminated frame (todo: only assert this if CRC correct); the frame includes 4 bytes CRC; rxfin is asserted during the second cycle after the CRC

       txbyte,
       txclk,
       txen,   // set to 1 to start transmission; must be set to 0 together after the last byte
       txact,  // is 1 when actually data bytes are being sent, but starting 1 cycle early to allow for pipelining
    );
    input clk; 
    output[7:0] rxbyte=rxd[7:0];
    output rxclk;
    output rxdv;
    output rxfin;
    input [7:0] txbyte;
    output txclk;
    input txen;
    output txact;


    wire tx_pclk, rx_pclk;
    wire rxclk = rx_pclk;
    wire txclk = tx_pclk;

    reg rst=1;     // power-on reset signal for just about everything in the SERDES; I'm surprised this is needed...
    wire tx_pcs_rst = rst, rx_ser_rst = rst, tx_ser_rst = rst, dual_rst = rst, serdes_dual_rst = rst;
    wire rx_los_lol, rx_cdr_lol;

    // --------- set up the SERDES ---------
    // largely based on Dave Shah's demo file, https://gist.githubusercontent.com/daveshah1/8ce26fc593bd837dcfa8b9cd28ee195b/raw/09dc5c40493128934bdc00eca5d5c4d980120d80/demo.v

    // note, using channel CH1 of the X69/Y71 DCU here, as that fit my layout best
    // the other three SERDESs in the '45 chip can be found by substituting CH0 and at X42/Y71

    (* BEL="X69/Y71/DCU" *)    // this looks like some kind of weird comment, but apparently chooses which physical DCU is to be used!
    DCUA DCU0_inst (
        .CH1_RX_REFCLK(clk),
        .CH1_FF_RXI_CLK(rx_pclk), .CH1_FF_RX_PCLK(rx_pclk),
        .CH1_FF_TXI_CLK(tx_pclk), .CH1_FF_TX_PCLK(tx_pclk),
        .CH1_FF_TX_D_0(txd[0]), .CH1_FF_TX_D_1(txd[1]),  .CH1_FF_TX_D_2(txd[2]),  .CH1_FF_TX_D_3(txd[3]),  
        .CH1_FF_TX_D_4(txd[4]), .CH1_FF_TX_D_5(txd[5]),  .CH1_FF_TX_D_6(txd[6]),  .CH1_FF_TX_D_7(txd[7]),  
        .CH1_FF_TX_D_8(txd[8]),  .CH1_FF_TX_D_9(1'b0)  ,  .CH1_FF_TX_D_10(1'b0),   .CH0_FF_TX_D_11(1'b0),
        .CH1_FFC_EI_EN(1'b0), .CH1_FFC_SIGNAL_DETECT(1'b0), .CH1_FFC_LANE_TX_RST(tx_pcs_rst), .CH1_FFC_LANE_RX_RST(rx_pcs_rst),
        .CH1_FFC_RRST(rx_ser_rst), .CH1_FFC_TXPWDNB(1), .CH1_FFC_RXPWDNB(1), .D_FFC_DUAL_RST(dual_rst),
        .CH1_FF_RX_D_0(rxd[0]), .CH1_FF_RX_D_1(rxd[1]), .CH1_FF_RX_D_2(rxd[2]), .CH1_FF_RX_D_3(rxd[3]),
        .CH1_FF_RX_D_4(rxd[4]), .CH1_FF_RX_D_5(rxd[5]), .CH1_FF_RX_D_6(rxd[6]), .CH1_FF_RX_D_7(rxd[7]),
        .CH1_FF_RX_D_8(rxd[8]),
        .CH1_FFS_RLOS(rx_los_lol), .CH1_FFS_RLOL(rx_cdr_lol),

        .D_FFC_MACRO_RST(serdes_dual_rst), .D_FFC_MACROPDB(1), .D_FFC_TRST(tx_ser_rst),
//        .D_REFCLKI(clk),    // this clock input becomes reference for the TX - I don't understand why this doesn't work, is my crystal too far off?
        .D_REFCLKI(rx_pclk),    // this clock input becomes reference for the TX - recycling the other side's clock; works fine, as long as only one side does it ;-)

        .CH1_FFC_ENABLE_CGALIGN(0),
        .CH1_FFC_CDR_EN_BITSLIP(0),
    );
                                                 // comments added based on documentation, experimentation, and guessing...    
                                                 // most values unchanged from Dave Shah's demo file
    defparam DCU0_inst.D_MACROPDB = "0b1";
    defparam DCU0_inst.D_IB_PWDNB = "0b1";
    defparam DCU0_inst.D_XGE_MODE = "0b0";
    defparam DCU0_inst.D_LOW_MARK = "0d4";
    defparam DCU0_inst.D_HIGH_MARK = "0d12";
    defparam DCU0_inst.D_BUS8BIT_SEL = "0b0";
    defparam DCU0_inst.D_CDR_LOL_SET = "0b00";    // LOL = Loss Of Lock; TN1261 p86 apparently the tolerance for detecting loss of lock, 300...4000 ppm
    defparam DCU0_inst.D_TXPLL_PWDNB = "0b1";     // TN1261 p87: 1 = power up
    defparam DCU0_inst.D_BITCLK_LOCAL_EN = "0b1";
    defparam DCU0_inst.D_BITCLK_ND_EN = "0b0";
    defparam DCU0_inst.D_BITCLK_FROM_ND_EN = "0b0";
    defparam DCU0_inst.D_SYNC_LOCAL_EN = "0b1";
    defparam DCU0_inst.D_SYNC_ND_EN = "0b0";
    defparam DCU0_inst.CH1_UC_MODE = "0b0";    // "User Configured" PCS mode (TN1261 p.90)
    defparam DCU0_inst.CH1_PCIE_MODE = "0b0";
    defparam DCU0_inst.CH1_RIO_MODE = "0b0";   // "Rapid IO" PCS mode
    defparam DCU0_inst.CH1_WA_MODE = "0b1";    // "Word Alignment" PCS mode; "1=bitslip, 0=barrel shift"
    defparam DCU0_inst.CH1_INVERT_RX = "0b0";
    defparam DCU0_inst.CH1_INVERT_TX = "0b0";
    defparam DCU0_inst.CH1_PRBS_SELECTION = "0b0";
    defparam DCU0_inst.CH1_GE_AN_ENABLE = "0b0";    // "1=Enable GigE Auto Negotiation" ... whatever that may mean, as the autonegotiation is definitely something that should be done in user logic
    defparam DCU0_inst.CH1_PRBS_LOCK = "0b0";
    defparam DCU0_inst.CH1_PRBS_ENABLE = "0b0";
    defparam DCU0_inst.CH1_ENABLE_CG_ALIGN = "0b1";  // "only in uc_mode - 1=enable continuous comma alignment"
    defparam DCU0_inst.CH1_TX_GEAR_MODE = "0b0";
    defparam DCU0_inst.CH1_RX_GEAR_MODE = "0b0";
    defparam DCU0_inst.CH1_PCS_DET_TIME_SEL = "0b00";  // "PCS connection detection time"
    defparam DCU0_inst.CH1_PCIE_EI_EN = "0b0";
    defparam DCU0_inst.CH1_TX_GEAR_BYPASS = "0b0";
    defparam DCU0_inst.CH1_ENC_BYPASS = "0b0";      // PCS: "Bypass 8b10b encoder"
    defparam DCU0_inst.CH1_SB_BYPASS = "0b0";       // PCS: "Invert TX data after SerDes Bridge"
    defparam DCU0_inst.CH1_RX_SB_BYPASS = "0b0";    // PCS: "Invert RX data after SerDesBridge"
    defparam DCU0_inst.CH1_WA_BYPASS = "0b0";       // PCS: "Bypass word alignment"
    defparam DCU0_inst.CH1_DEC_BYPASS = "0b0";      // PCS: "Bypass 8b10b decoder"
    defparam DCU0_inst.CH1_CTC_BYPASS = "0b1";      // CTC = Clock Tolerance Compensation
    defparam DCU0_inst.CH1_RX_GEAR_BYPASS = "0b0";  // "1=Bypass PCS Rx gear box"
    defparam DCU0_inst.CH1_LSM_DISABLE = "0b0";     // PCS: "1=Disable Rx link state machine" ... but enabling it doesn't seem to do much, I still need to force a resync from user logic
    defparam DCU0_inst.CH1_MIN_IPG_CNT = "0b11";    // PCS: "Minimum IPG to enforce", inter packet gap
    defparam DCU0_inst.CH1_UDF_COMMA_MASK = "0x3ff";   // RXCOMMAM in TN1261
    defparam DCU0_inst.CH1_UDF_COMMA_A = "0x283";
    defparam DCU0_inst.CH1_UDF_COMMA_B = "0x17C";

    // related to TX clock: ; bitrate = refclk * {8,10,16,20,25} / {1,2,11}
    defparam DCU0_inst.D_REFCK_MODE = "0b000";   // presumably the MSB is REFCK25X; then 000..100 = internal high speed bit clock is 20x, 10x, 16x, 8x, 25x.
       // this affects both TX and RX
    defparam DCU0_inst.CH1_TX_DIV11_SEL = "0b0";    // select divide by 11 transmit clock
    defparam DCU0_inst.CH1_RATE_MODE_TX = "0b1";    // Serdes: "0/1 = full/half rate selection for transmit"

    // related to RX clock:
    defparam DCU0_inst.CH1_RX_DCO_CK_DIV = "0b000";  // Serdes: "VCO output frequency select; 00x div by1, 01x div by2, 100...111 div by 4/8/16/32"
    defparam DCU0_inst.CH1_RATE_MODE_RX = "0b1";     // 1 = divide by 2 (if DIV11==0) "0/1 = full/half rate selection for receive"
    //each of the above two can be set to divide by 2, but they don't seem to have the same effect on the system: rx_pclk is the same, but recovered data is wrong in one case
    defparam DCU0_inst.CH1_RX_DIV11_SEL = "0b0";     // 1 = divide by 11
    defparam DCU0_inst.CH1_RX_RATE_SEL = "0d00";     // may not be clock-related at all; doc suggests it's about equalizer settings...
    defparam DCU0_inst.CH1_DCOIOSTUNE = "0b100";     // was 010; setting to 100 seems to allow slightly lower rx clock freq, good for 1.25 Gbit/s
    defparam DCU0_inst.CH1_DCOITUNE = "0b10";        // was 00; setting to 10, combined with DCOODSTUNE=100, allows down to 1 Gbit/s
    defparam DCU0_inst.CH1_DCOITUNE4LSB = "0b010";   // there are more DCO related settings further down the list

    defparam DCU0_inst.CH1_RCV_DCC_EN = "0b0";        // Serdes: "1=Receiver DC coupling enable; 0=AC coupling (default)"
    defparam DCU0_inst.CH1_TPWDNB = "0b1";            // TN1261 p94: 1 = power up
    defparam DCU0_inst.CH1_RTERM_TX = "0d19";         // Serdes: "TX resistor termination select" dec19 = 50 ohm
    defparam DCU0_inst.CH1_TX_CM_SEL = "0b00";        // Serdes: "TX output common mode voltage 00=power down; 01/10/11 = 0.6/0.55/0.5 V"
    defparam DCU0_inst.CH1_TDRV_PRE_EN = "0b0";       // Serdes: "TX driver de-emphasis enable"
    defparam DCU0_inst.CH1_TDRV_SLICE0_SEL = "0b01";  // Serdes: "TX drive slice enable for slice 0"
    defparam DCU0_inst.CH1_TDRV_SLICE1_SEL = "0b00";  // etc; 00=power down; 01=select main data;
    defparam DCU0_inst.CH1_TDRV_SLICE2_SEL = "0b01";  // 10 = select pre data; 11 = select post data"
    defparam DCU0_inst.CH1_TDRV_SLICE3_SEL = "0b01";  // together these and the following apparently determine both the magnitude and the pulse form of the transmit signal
    defparam DCU0_inst.CH1_TDRV_SLICE4_SEL = "0b00";
    defparam DCU0_inst.CH1_TDRV_SLICE5_SEL = "0b00";
    defparam DCU0_inst.CH1_TDRV_SLICE0_CUR = "0b011";  // 000...111 = 100...800 uA
    defparam DCU0_inst.CH1_TDRV_SLICE1_CUR = "0b000";
    defparam DCU0_inst.CH1_TDRV_SLICE2_CUR = "0b11";
    defparam DCU0_inst.CH1_TDRV_SLICE3_CUR = "0b10";  // Serdes: "Controls the output current for slice 4. Every 100 uA increases the output amplitude by 100 mV"
    defparam DCU0_inst.CH1_TDRV_SLICE4_CUR = "0b00";  // " 00/01/10/11 = 800/1600/2400/3200 uA"
    defparam DCU0_inst.CH1_TDRV_SLICE5_CUR = "0b00";
    defparam DCU0_inst.CH1_TDRV_DAT_SEL = "0b00";     // "Driver output select; 00=data from serializer muxed to driver (normal operation)"
    defparam DCU0_inst.CH1_RPWDNB = "0b1";
    defparam DCU0_inst.CH1_SEL_SD_RX_CLK = "0b1";
    defparam DCU0_inst.CH1_FF_RX_H_CLK_EN = "0b0";    // "FF" = "FPGA Fabric"
    defparam DCU0_inst.CH1_FF_RX_F_CLK_DIS = "0b0";
    defparam DCU0_inst.CH1_FF_TX_H_CLK_EN = "0b0";
    defparam DCU0_inst.CH1_FF_TX_F_CLK_DIS = "0b0";
    defparam DCU0_inst.CH1_TDRV_POST_EN = "0b0";   // "1=TX driver post-emphasis enable"
    defparam DCU0_inst.CH1_TX_POST_SIGN = "0b0";   // "1=TX post emphasis not inverted"
    defparam DCU0_inst.CH1_TX_PRE_SIGN = "0b0";   // "1=TX preemphasis not inverted"
    defparam DCU0_inst.CH1_REQ_LVL_SET = "0b00";  // "Level setting for equalization; 00/01/10=6/9/12dB; 11=not used"
    defparam DCU0_inst.CH1_REQ_EN = "0b1";        // "1=Receiver equalization enable"
    defparam DCU0_inst.CH1_RTERM_RX = "0d22";     // "RX input termination setting; dec22=reserved"(?!)
    defparam DCU0_inst.CH1_RXTERM_CM = "0b11";    // "command[sic] mode voltage for RX input termination"
    defparam DCU0_inst.CH1_PDEN_SEL = "0b1";
    defparam DCU0_inst.CH1_RXIN_CM = "0b11";      // "common mode voltage for equalizer input in ac coupling"
    defparam DCU0_inst.CH1_LEQ_OFFSET_SEL = "0b0";      // LEQ = Line EQualizer
    defparam DCU0_inst.CH1_LEQ_OFFSET_TRIM = "0b000";
    defparam DCU0_inst.CH1_RLOS_SEL = "0b1";      // "enables LOS detector output before .... (if the channel is being used, this bit should be set to 1)"
    defparam DCU0_inst.CH1_RX_LOS_LVL = "0b100";    // LOS = Loss Of Signal
    defparam DCU0_inst.CH1_RX_LOS_CEQ = "0b11";
    defparam DCU0_inst.CH1_RX_LOS_HYST_EN = "0b0";
    defparam DCU0_inst.CH1_RX_LOS_EN = "0b1";
    defparam DCU0_inst.CH1_LDR_RX2CORE_SEL = "0b0";    // LDR = Low Data Rate, process data directly in FPGA logic, bypassing the SERDES
    defparam DCU0_inst.CH1_LDR_CORE2TX_SEL = "0b0";
    defparam DCU0_inst.D_ISETLOS = "0d0";
    defparam DCU0_inst.D_SETIRPOLY_AUX = "0b10";
    defparam DCU0_inst.D_SETICONST_AUX = "0b01";
    defparam DCU0_inst.D_SETIRPOLY_CH = "0b10";
    defparam DCU0_inst.D_SETICONST_CH = "0b10";
    defparam DCU0_inst.D_REQ_ISET = "0b001";
    defparam DCU0_inst.D_PD_ISET = "0b00";
    defparam DCU0_inst.D_DCO_CALIB_TIME_SEL = "0b00";
    defparam DCU0_inst.CH1_CDR_CNT4SEL = "0b00";
    defparam DCU0_inst.CH1_CDR_CNT8SEL = "0b00";
    defparam DCU0_inst.CH1_DCOATDCFG = "0b00";       // DCO seems to be the oscillator used for RX clock, although the manual doesn't expand this abbreviation
    defparam DCU0_inst.CH1_DCOATDDLY = "0b00";
    defparam DCU0_inst.CH1_DCOBYPSATD = "0b1";
    defparam DCU0_inst.CH1_DCOCALDIV = "0b100";
    defparam DCU0_inst.CH1_DCOCTLGI = "0b011";
    defparam DCU0_inst.CH1_DCODISBDAVOID = "0b0";
    defparam DCU0_inst.CH1_DCOFLTDAC = "0b00";
    defparam DCU0_inst.CH1_DCOFTNRG = "0b001";
    defparam DCU0_inst.CH1_DCOIUPDNX2 = "0b1";
    defparam DCU0_inst.CH1_DCONUOFLSB = "0b100";
    defparam DCU0_inst.CH1_DCOSCALEI = "0b01";
    defparam DCU0_inst.CH1_DCOSTARTVAL = "0b010";
    defparam DCU0_inst.CH1_DCOSTEP = "0b11";
    defparam DCU0_inst.CH1_BAND_THRESHOLD = "0d0";
    defparam DCU0_inst.CH1_AUTO_FACQ_EN = "0b1";     // FACQ = Frequency ACQuition process
    defparam DCU0_inst.CH1_AUTO_CALIB_EN = "0b1";
    defparam DCU0_inst.CH1_CALIB_CK_MODE = "0b0";
    defparam DCU0_inst.CH1_REG_BAND_OFFSET = "0d0";
    defparam DCU0_inst.CH1_REG_BAND_SEL = "0d0";
    defparam DCU0_inst.CH1_REG_IDAC_SEL = "0d0";
    defparam DCU0_inst.CH1_REG_IDAC_EN = "0b0";
    defparam DCU0_inst.D_CMUSETISCL4VCO = "0b000";    // could CMU mean Clock Multiplier Unit? But these are far too many parameters to play with...
    defparam DCU0_inst.D_CMUSETI4VCO = "0b00";
    defparam DCU0_inst.D_CMUSETINITVCT = "0b00";
    defparam DCU0_inst.D_CMUSETZGM = "0b000";
    defparam DCU0_inst.D_CMUSETP2AGM = "0b000";
    defparam DCU0_inst.D_CMUSETP1GM = "0b000";
    defparam DCU0_inst.D_CMUSETI4CPZ = "0d3";
    defparam DCU0_inst.D_CMUSETI4CPP = "0d3";
    defparam DCU0_inst.D_CMUSETICP4Z = "0b101";
    defparam DCU0_inst.D_CMUSETICP4P = "0b01";
    defparam DCU0_inst.D_CMUSETBIASI = "0b00";
    defparam DCU0_inst.D_SETPLLRC = "0d1";  // undocumented; doesn't seem to affect the RXPLL's frequency range, perhaps only its loopbandwidth?
    defparam DCU0_inst.D_TX_VCO_CK_DIV = "0b000";
    defparam DCU0_inst.D_PLL_LOL_SET = "0b00";    // LOL = Loss Of Lock
    defparam DCU0_inst.D_RG_EN = "0b0";
    defparam DCU0_inst.D_RG_SET = "0b00";






    // ----------- transmit logic for gigabit ethernet over fiber (this is far from a full IEEE802-compliant implementation!)

    reg [2:0] ctr;  // counter running at txclk
    reg [3:0] txphase=0;
    wire txodd = ~ctr[0];
    reg [31:0] txcrc;
    reg txact=0;
    reg [7:0] txbyte2;
    reg txen1;
    reg [1:0] autoneg1;
    reg [1:0] autoneg2;
    reg [8:0] tmp8;

    // take care of the autonegotiation:
    wire [8:0] txd = 
       (autoneg2==3) ? tmp8 :     // if autonegotiation finished
       // otherwise, transmit autonegotiation pattern depending on what the other side has already sent
       // (this naive implementation counts on the other side having a more complete standards-compliant implementation, as we just follow their actions)
       (ctr[1:0]==0) ? 9'h1bc :
       (ctr[1:0]==1) ? (ctr[2] ? 9'h042 : 9'h0b5) :
       (autoneg2==0) ? 9'h000 : (ctr[1:0]==2) ? 9'h020 : 
       (autoneg2[1] ? 9'h000 : 9'h040);  // we advertise only fullduplex, no pause
    
    // if autonegotiation finished, this statemachine handles frame transmission:
    always @(posedge tx_pclk) begin
        if (ctr[2:0]==7) begin
           autoneg2 <= autoneg1;
           autoneg1 <= autoneg0;
        end
        ctr <= ctr + 1;
        rst <= 0;
        txbyte2 <= txbyte;
        txen1 <= txen;
        case (txphase)
            0: begin
                   tmp8 <= txodd ? 9'h050 : 9'h1bc;  // idle
                   if (txodd && txen1) txphase <= txphase+1;
                end
            1: begin
                   tmp8 <= txodd ? 9'h055 : 9'h1fb;  // start-of-frame
                   if (txodd) txphase <= txphase+1;
                   txcrc<=32'b11111111111111111111111111111111;
                end
            2, 3, 4: begin
                   tmp8 <= (txodd && txphase==4) ? 9'h0d5 : 9'h055;  // preamble, last byte is different
                   if (txodd) txphase <= txphase+1;
                   if (txodd && txphase==3) txact <= 1;    // txact is enabled "early" for pipelining
                end
            5: begin                                
                   if (txen1) begin
                      tmp8 <= { 1'b0, txbyte2[7:0] };   // contents of the frame
                      txcrc <= nextCRC32_D8( { txbyte2[0], txbyte2[1], txbyte2[2], txbyte2[3], txbyte2[4], txbyte2[5], txbyte2[6], txbyte2[7] } , txcrc);
                   end else begin
                      txphase <= txphase+1;             // end of data, start transmitting the CRC
                      tmp8 <= ~{ 1'b1, txcrc[24], txcrc[25], txcrc[26], txcrc[27], txcrc[28], txcrc[29], txcrc[30], txcrc[31] };
                      txcrc[31:8]<=txcrc[23:0];
                   end
                end
            6, 7, 8: begin                              // transmit CRC
                   tmp8 <= ~{ 1'b1, txcrc[24], txcrc[25], txcrc[26], txcrc[27], txcrc[28], txcrc[29], txcrc[30], txcrc[31] };
                   txcrc[31:8]<=txcrc[23:0];
                   txphase <= txphase+1;
                end
            9: begin
                   tmp8 <= 9'h1fd;          // end-of-frame
                   txphase <= txphase+1;
                end
            10: begin
                   tmp8 <= 9'h1f7;          // carrier extend, 1 or 2 symbols, to return to the original even/odd rhythm
                   if (txodd) begin
                      txphase <= 0;         // go back to idle state
                      txact <= 0;
                   end
                end
        endcase
    end


    // ----------- receive logic for gigabit ethernet over fiber (this is far from a full IEEE802-compliant implementation!)

    wire [8:0] rxd;
    reg [12:0] commawatchdog;
    wire rx_pcs_rst=&commawatchdog;    // reset the byte synchronization if we don't see a valid control symbol for a while (the IEEE802 doc specifies a much more detailed procedure for this...)
    reg [2:0] rxstate=1;
    reg [1:0] autoneg=0;   // 0=nothing seen yet; 1 = received partner's info; 2 = received partner's ack; 3 = partner no longer sending config frames, ready for data
    reg [1:0] autoneg0=0;

    always @(posedge rx_pclk) begin

       if (rxd[8] && rxd[7:0]!=8'hee) commawatchdog<=0;   // resets the watchdog on any valid control symbol, not just on actual commas
       else commawatchdog <= commawatchdog+1;

       if (rx_pcs_rst) begin
          // watchdog barks
          rxstate<=0;
          autoneg<=0;
       end else begin
          // normal situation
          case (rxstate)
             0: begin
                 if (rxd[8:0]==9'h1fb) rxstate<=1;   // idle; wait for start-of-packet
                 if (rxd[8:0]==9'h1bc) rxstate<=4;   // or for K28.5
                 // anything else gets skipped; could be D5.6 or D16.2 for real idle, or K23.7 carrier_extend
                end
             1: begin
                 if (rxd[7:0]==8'hd5) rxstate<=2;    // wait for end of preamble
                 if (rxd[8]==1) rxstate<=0;          // control symbol shouldn't happen here in a normal frame, but we shouldn't be stuck in this state if the other side for some reason doesn't terminate the frame properly
                end
             2: begin
                 if (rxd[8:0]==9'h1fd) rxstate<=3;   // wait for end-of-packet (K29.7)
                 else if (rxd[8]==1) rxstate<=0;     // other control symbol shouldn't happen in a normal frame - cancel it
                end
             3: rxstate<=0;                          // frame ends; only purpose of this state is to produce the rxfin signal
             4: begin      // check the symbol after a K28.5, to see whether it's perhaps a configuration message
                 if (rxd[8:0]==9'h42 || rxd[8:0]==9'hb5) rxstate<=5;
                 else begin
                    rxstate<=0;
                    autoneg<=3;
                 end
                end
             5: rxstate<=6;  // skip first config byte;
             6: begin
                 if (rxd[6]) autoneg<=2;   // if the ack bit is set
                 else autoneg<=1;          // otherwise
                 rxstate<=0;  // this was the second config byte, back to idle
                end
          endcase;
       end
       autoneg0<=autoneg;
    end
    wire rxdv = (rxstate==2);
    wire rxfin = (rxstate==3);

endmodule

